/* ESP-01-LCD-RTC-NTP-Clock.ino
 * by Cyberblogspot 09Feb2020
 * 
 * NTP code is from Time Library:
 * TimeNTP_ESP8266WiFi.ino
 * Example showing time sync to NTP time source
 */

#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <TimeLib.h>
#include <RTClib.h>
#include <LiquidCrystal_I2C.h>

const char ssid[] = "LVM4G";  //  your network SSID (name)
const char pass[] = "filislands";  // your network password

// NTP Servers:
static const char ntpServerName[] = "ntp.pagasa.dost.gov.ph"; 
//static const char ntpServerName[] = "us.pool.ntp.org";
//static const char ntpServerName[] = "time.nist.gov";
//static const char ntpServerName[] = "time-a.timefreq.bldrdoc.gov";
//static const char ntpServerName[] = "time-b.timefreq.bldrdoc.gov";
//static const char ntpServerName[] = "time-c.timefreq.bldrdoc.gov";

const int timeZone = 8; 
//const int timeZond = 1;   // Central European Time
//const int timeZone = -5;  // Eastern Standard Time (USA)
//const int timeZone = -4;  // Eastern Daylight Time (USA)
//const int timeZone = -8;  // Pacific Standard Time (USA)
//const int timeZone = -7;  // Pacific Daylight Time (USA)

WiFiUDP Udp;
unsigned int localPort = 8888;  // local port to listen for UDP packets

LiquidCrystal_I2C lcd(0x27, 20, 4);
RTC_DS3231 rtc;

time_t lastNtp;  // last successful NTP sync
#define ntpInterval 60 //for testing 60 seconds
                       //could be daily or weekly as we're using  
                       //a precise rtc clock module 
void setup() {
  lcd.begin(0, 2);  // GPIO-0 SDA, GPIO-2 SCL
  lcd.backlight();
  lcd.clear();
   
  //Serial.begin(115200);
  
  syncRtc();
    
  setSyncProvider(rtcGet);
  setSyncInterval(60); // 1 min for testing (default 300 - 5min)  
}

void loop() {
  static time_t tLast;
  int h, m, s, y, mo, d, dow;

  time_t tNow = now();

  if (tNow != tLast){
    tLast = tNow;
    
    //h = hour(tNow); // 24-hour format
    h = hourFormat12(tNow);
    m = minute(tNow);
    s = second(tNow);
    y = year(tNow);
    mo = month(tNow);
    d = day(tNow);
    dow = weekday(tNow);

    // Print time on first line
    lcd.setCursor(1, 0);
    lcd.print("TIME: ");
    if (h < 10) lcd.print('0');
    lcd.print(h);
    lcd.print(':');
    if (m < 10) lcd.print('0');
    lcd.print(m);
    lcd.print(':');
    if (s < 10) lcd.print('0');
    lcd.print(s); 
    lcd.print(' ');
    if (isAM()) lcd.print("AM");
      else lcd.print("PM");  

    //Print date on second line
    lcd.setCursor(2, 1);
    lcd.print(dayShortStr(dow));
    lcd.print(' ');
    if (d < 10) lcd.print('0');
    lcd.print(d);
    lcd.print(' ');
    lcd.print(monthShortStr(mo)); 
    lcd.print(' ');
    lcd.print(y); 

    //Print temperature on third line three
    lcd.setCursor(0, 2);
    lcd.print("Temp: ");
    float tempC = rtc.getTemperature();
    float tempF = tempC * 9. / 5. + 32.;
    lcd.print(tempC);
    lcd.print('C');
    lcd.print("  ");
    lcd.print(tempF);
    lcd.print('F');    
  }
    
  if (tNow >= lastNtp + ntpInterval){   
    syncRtc();  // sync RTC with NTP 
  }
}

/*
 * void syncRtc()
 * 
 * Connects to the internet in non-blocking mode
 * 
 * Once connected, gets the NTP time, synchronizes the RTC clock,
 * and disconnects from the internet
 * 
 * Displays the last successfull NTP sync date and time
 * Prints NOK if RTC needs sync or OK after a successful sync
 * 
 */
 
bool pendingConnect = false;

void syncRtc(){  
  if (!pendingConnect){
    WiFi.begin(ssid, pass);
    pendingConnect = true;
  } 
  
  if (WiFi.status() == WL_CONNECTED){
    pendingConnect = false;
  
    Udp.begin(localPort);

    time_t tNow = getNtpTime();  //returns 0 on fail

    if (tNow != 0) {
      rtc.adjust(DateTime(tNow));
      lastNtp = tNow;
      
      lcd.setCursor(0, 3);
      lcd.print("OK ");

      // Display lastNtP on fourth line of LCD
      int y, mo, d, h, m, s;
      y = year(lastNtp)-2000;
      mo = month(lastNtp);
      d = day(lastNtp);
      h = hour(lastNtp);
      m = minute(lastNtp);
      s = second(lastNtp);
      
      lcd.setCursor(3, 3);
      lcd.print(y);
      lcd.print('.');
      if (mo < 10) lcd.print('0');
      lcd.print(mo);
      lcd.print('.');
      if (d < 10) lcd.print('0');
      lcd.print(d);
      lcd.print('.');
      if (h < 10) lcd.print('0');
      lcd.print(h);
      lcd.print('.');
      if (m < 10) lcd.print('0');
      lcd.print(m);
      lcd.print('.');
      if (s < 10) lcd.print('0');
      lcd.print(s);

      // For battery operation, test ESP8266 sleep mode
      WiFi.disconnect();
      WiFi.mode(WIFI_OFF);   
    }  
  }
  else {
    lcd.setCursor(0,3);  // not yet connected, exit promptly
    lcd.print("NOK");    // we'll try on the next loop
  }
}

// Sync provider function
time_t rtcGet(){
  return rtc.now().unixtime();
}
 
/*-------- NTP code ----------*/

const int NTP_PACKET_SIZE = 48; // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming & outgoing packets

time_t getNtpTime()
{ 
  IPAddress ntpServerIP; // NTP server's ip address

  while (Udp.parsePacket() > 0) ; // discard any previously received packets
  //Serial.println("Transmit NTP Request");
  // get a random server from the pool
  WiFi.hostByName(ntpServerName, ntpServerIP);
  //Serial.print(ntpServerName);
  //Serial.print(": ");
  //Serial.println(ntpServerIP);
  sendNTPpacket(ntpServerIP);
  uint32_t beginWait = millis();
  while (millis() - beginWait < 1500) {
    int size = Udp.parsePacket();
    if (size >= NTP_PACKET_SIZE) {
      //Serial.println("Receive NTP Response");
      Udp.read(packetBuffer, NTP_PACKET_SIZE);  // read packet into the buffer
      unsigned long secsSince1900;
      // convert four bytes starting at location 40 to a long integer
      secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
      secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
      secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
      secsSince1900 |= (unsigned long)packetBuffer[43];
      
      return secsSince1900 - 2208988800UL + timeZone * SECS_PER_HOUR;
    }
  }
  return 0; // return 0 if unable to get the time
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address)
{
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123); //NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}
